// Handle reading of texts in linear order.
//
// $Id: read-order.cc,v 1.15 1994/07/30 22:11:20 inge Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>

#include "conference.h"
#include "connection.h"
#include "get_membership_q.h"
#include "kom-types.h"
#include "maps.h"
#include "membership.h"
#include "misc.h"
#include "read-order.h"
#include "text-mapping.h"
#include "text.h"

// ================================================================ 
//                    Constructors and Destructors                  


Read_order::Read_order(Session *sess, Conf_no cno, Pers_no reader)
    : session(sess), persons(sess->persons()), 
      conferences(sess->conferences()), map(sess->maps()[cno]), 
      texts(sess->texts()),
      conf_no(cno), pers_no(reader), pending_membership_question(NULL),
      membership(),
      unread_texts(), first_unchecked(0)
{
}
 

//Read_order::Read_order(const Read_order &ro)
//    : conferences(ro.conferences), map(ro.map),
//      conf_no(ro.conf_no), pers_no(ro.pers_no),
//      unread_texts(ro.unread_texts), first_unchecked(ro.first_unchecked)
//{
//}


Read_order::~Read_order()
{
    delete pending_membership_question;
}


// ================================================================ 
//

//const Read_order & 
//Read_order::operator = (const Read_order &)
//{
//    cerr << "Operator= for class Read_order is not yet implemented\n";
//    return *this;
//}


// ================================================================


int
Read_order::prefetch_map(int texts)
{
    Conference conf = conferences[conf_no];
    if (!membership.valid() && pending_membership_question == NULL)
    {
	pending_membership_question = 
	    new get_membership_question(conferences.connection(), 
					pers_no, conf_no);
    }

    switch (conf.prefetch())
    {
    case st_pending:
	// We can't do anything sensible unless we have the confstat.
	return 1 + (pending_membership_question != NULL
		    && pending_membership_question->status() == st_pending);

    case st_error:
	// Now what? +++FIXME///
	// (Note: This case CAN occur if the conference is being
	// deleted during a session).
	return 0;

    case st_ok:
        {
	    Local_text_no first_local = conf.first_local_no();
	    Local_text_no limit = first_local + conf.no_of_texts();

	    if (unread_texts.length() < texts && first_unchecked < limit)
	    {
		if (pending_membership_question != NULL)
		{
		    switch (pending_membership_question->status())
		    {
		    case st_pending:
			return 1;
			
		    case st_error:
			// Now what? +++FIXME///
			return 0;
		    
		    case st_ok:
			receive_membership();
			break;
		    
		    default:
			abort();
		    }
		}

		if (first_unchecked < first_local)
		    first_unchecked = first_local;

		while (first_unchecked < limit 
		       && unread_texts.length() < texts)
		{
		    if (!membership.is_read(first_unchecked))
		    {
			// FIXME+++: The "50" on the next line used to be
			// "texts", but that caused a very large question
			// to be issued. The map question should be split
			// into severaly small chunks, but I'm not sure
			// where the splitting should be done.
		        // Temporarily, do it this way.
			
			if (map.prefetch(first_unchecked, 50) == st_pending)
			    return 1;
			
			Text_no tno = map[first_unchecked];
			if (tno != 0)
			    unread_texts.append(tno);
		    }
		    
		    first_unchecked++;
		}
	    }
	}
	break;

    default:
	abort();
    }

    return 0;
}

int
Read_order::prefetch(int required_texts,
		     int max_question,
		     int min_question)
{
    int questions = prefetch_map(required_texts);
    
    for (Pix ix = unread_texts.first();
	 ix != 0 && ((required_texts > 0 && questions < max_question)
		     || questions < min_question);
	 unread_texts.next(ix))
    {
	Text t = texts[unread_texts(ix)];
	switch (t.prefetch())
	{
	case st_pending:
	    questions++;
	    break;
	    
	case st_error:
	    // This is strange. The text must have been deleted very
	    // resently. Remove it. (This may cause some problems if
	    // this is the text we are about to show).
	    unread_texts.del(ix, -1);
	    break;

	case st_ok:
	    // Go ahead and prefetch some other stuff...
	    {
		int nq = t.prefetch_for_display(required_texts > 0 ? 
						max_question - questions : 
						min_question - questions,
						session);
		questions += nq;
		if (nq == 0)
		    required_texts--;
	    }
	    break;

	default:
	    abort();
	}
    }

    return questions;
}

			
Prompt
Read_order::next_prompt()
{
    while (prefetch_map(1) > 0)
    {	
	switch (Connection::drain(100))
	{
	case st_pending:
	    return TIMEOUT_PROMPT;
	case st_error:
	    return PROMPT_ERROR;
	case st_ok:
	    break;
	}
    }

    if (unread_texts.empty())
	return NO_MORE_TEXT;
    else
	return NEXT_TEXT;
}


//
// Return the next text to be read within this conference.
// Return 0 if there is no more unread text.
//

Text_no
Read_order::next_text(Bool remove)
{
    while (prefetch_map(1) > 0)
	if (Connection::drain(100) == st_error)
	    return 0;

    if (unread_texts.empty())
	return 0;

    if (remove)
	return unread_texts.remove_front();
    else
	return unread_texts.front();
}


Bool
Read_order::remove_text(Text_no tno)
{
    // FIXME: We would like to make sure that the information up to tno
    //        is fetched before we do this.
    // FIXME: If we use the local textno instead of the global we will 
    //        probably save some work.

    if (unread_texts.empty())
	return TRUE;

    for (Pix ix = unread_texts.first(); ix != 0; unread_texts.next(ix)) {
	if (unread_texts(ix) == tno) {
	    unread_texts.del(ix);
	    break;
	}
    }

    return (Bool) unread_texts.empty();
}


Bool
Read_order::empty()
{
    while (prefetch_map(1) > 0
	   && Connection::drain(100) != st_error)
	continue;

    return (Bool) unread_texts.empty();
}


long
Read_order::min_num_unread()
{
    return unread_texts.length();
}

 
void
Read_order::receive_membership()
{
    membership = pending_membership_question->result();
    delete pending_membership_question;
    pending_membership_question = NULL;
}    

long
Read_order::max_num_unread()
{
    Conference conf = conferences[conf_no];
    if (!conf.present())
	return LONG_MAX;

    if (pending_membership_question != NULL
	&& pending_membership_question->status() == st_ok)
    {
	receive_membership();
    }

    if (!membership.valid())
    {
	// +++ We could subtract any texts here that we know are
	// deleted. Such information may exist in the text-mapping.
	return conf.no_of_texts();
    }

    long result = unread_texts.length();

    Local_text_no limit = conf.first_local_no() + conf.no_of_texts();

#if 0
    // This is SLOW when you have several thousand unread in I]M.
    for (Local_text_no ln = ((first_unchecked > membership->last_text_read())
			     ? first_unchecked : membership->last_text_read());
	 ln < limit; 
	 ln++)
    {
	if (!membership->is_read(ln))
	{
	    // +++ Should check if we know that this text is deleted.
	    result++;
	}
    }
#else
    if (first_unchecked > membership.last_text_read())
	result += limit - first_unchecked;
    else
	result += (limit - membership.last_text_read()
		   - membership.num_read_after_last_text_read());
#endif

    return result;    
}

Conf_no
Read_order::confno() const
{
    return conf_no;
}
